# -*- coding:utf-8 -*-
##############################################################
# Created Date: Tuesday, January 17th 2023
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################

# from utility_lib import func_running_time
import pandas as pd
# import sys
import os
from pathlib import Path

from utdf2gmns.utils_lib.package_settings import (utdf_link_column_names,
                                                  utdf_categories,
                                                  utdf_city_name)
from utdf2gmns.utils_lib.utility_lib import func_running_time, path2linux


# aviod the warning of "A value is trying to be set on a copy of a slice from a DataFrame"
pd.options.mode.chained_assignment = None  # default='warn'


@func_running_time
def read_UTDF_file(path_utdf: str) -> dict:
    """ read UTDF file and split data into different categories """

    # read the utdf.csv file
    with open(path_utdf, encoding='utf-8') as f:
        lines = f.readlines()

    # find the start index of each category, the index is the row contain column names
    categorical_data_beginning_index_dict = {}
    for i in range(len(lines)):
        if "Network" in lines[i] and utdf_categories["Network"] in lines[i + 1]:
            categorical_data_beginning_index_dict[i + 2] = "Network"
        elif "Nodes" in lines[i] and utdf_categories["Nodes"] in lines[i + 1]:
            categorical_data_beginning_index_dict[i + 2] = "Nodes"
        elif "Links" in lines[i] and utdf_categories["Links"] in lines[i + 1]:
            categorical_data_beginning_index_dict[i + 2] = "Links"
        elif "Lanes" in lines[i] and utdf_categories["Lanes"] in lines[i + 1]:
            categorical_data_beginning_index_dict[i + 2] = "Lanes"
        elif "Timeplans" in lines[i] and utdf_categories["Timeplans"] in lines[i + 1]:
            categorical_data_beginning_index_dict[i + 2] = "Timeplans"
        elif "Phases" in lines[i] and utdf_categories["Phases"] in lines[i + 1]:
            categorical_data_beginning_index_dict[i + 2] = "Phases"
        else:
            continue

    categorical_index_ordered = sorted(list(categorical_data_beginning_index_dict.keys())) # ascending order

    # prepare dataframe for each category
    utdf_dict_data = {}

    for j in range(len(categorical_index_ordered)):
        # get the category name from start_index_dict
        category_name = categorical_data_beginning_index_dict[categorical_index_ordered[j]]

        # if it's the last value in the list, then the end index is the end of the file
        if j == len(categorical_index_ordered) - 1:
            category_value = [k.split(",") for k in lines[categorical_index_ordered[j]:]]

        # if it's not the last value in the list, then the end index is the start index of the next category - 2
        else:
            category_value = [k.split(",") for k in lines[categorical_index_ordered[j]:categorical_index_ordered[j + 1] - 2]]

        # save data to dictionary
        utdf_dict_data[category_name] = pd.DataFrame(category_value[1:], columns=category_value[0])

    # format utdf_lane data : remove unnecessary rows with None and column with '\n'
    # format each table in utdf_dict_data
    for table_name, df_table_name in utdf_dict_data.items():
        try:
            # get the last column name from utdf_setting
            last_col_name = list(df_table_name.columns)[-1]

            # remove unnecessary rows / invalid rows with NaN
            if table_name != "Network":
                df_table_name = df_table_name[df_table_name["INTID"].notna()]
                df_table_name = df_table_name[df_table_name['INTID'].astype(str).str.isdigit()]
            else:
                df_table_name = df_table_name[df_table_name[last_col_name].notna()]

            # clean the data / remove '\n' in the end of column SW
            df_table_name.loc[:, last_col_name] = df_table_name[last_col_name].map(
                lambda x: x.replace("\n", ""))

            df_table_name = df_table_name.rename(columns={last_col_name: last_col_name.replace("\n", "")})

            # drop the column with empty string
            df_table_name = df_table_name.drop(columns=[""], axis=1) if "" in df_table_name.columns else df_table_name

            utdf_dict_data[table_name] = df_table_name
        except Exception as e:
            print(f"Could not format table: {table_name} for {e}")
            continue

    # update Timeplans table with three columns needed
    utdf_dict_data["Timeplans"] = utdf_dict_data["Timeplans"].iloc[:, 0:3]

    # utdf_dict_data["phase_timePlans"] = pd.concat([utdf_dict_data.get("Phases"), utdf_dict_data.get("Timeplans")], axis=0, ignore_index=True)
    utdf_dict_data["phase_timeplans"] = spanning_phase_timeplans_data(utdf_dict_data)

    return utdf_dict_data


@func_running_time
def generate_intersection_data_from_utdf(utdf_dict_data: dict, city_name: str) -> pd.DataFrame:

    # Get link data from utdf
    df_link = utdf_dict_data["Links"]

    # update columns name
    df_link.columns = [i.replace("\n", "") if "\n" in i else i for i in df_link.columns.tolist()]

    # remove unnecessary rows / invalid rows with NaN
    df_link = df_link[df_link["INTID"].notna()]

    # clean the data / remove '\n' in the end of column SW
    # df_link = df_link.rename(columns={"SW\n": "SW"})
    # df_link["SW"] = df_link["SW"].map(lambda x: x.replace("\n", ""))

    # get the unique link id
    link_id = df_link["INTID"].unique().tolist()

    # generate link dictionary in format of: {link_id: {RECORDNAME:{columns:values}}}
    df_link_dict = {}
    for single_id in link_id:
        df_single_id = df_link[df_link["INTID"] == single_id]
        record_name = df_single_id["RECORDNAME"].unique().tolist()
        df_single_id_dict = {}
        for name in record_name:
            # convert one row of dataframe to dictionary
            df_single_id_name = df_single_id[df_single_id["RECORDNAME"] == name].to_dict("records")[0]
            df_single_id_dict[name] = df_single_id_name
        df_link_dict[single_id] = df_single_id_dict

    # prepare intersection dataframe
    sequenced_intersection_id = 0
    link_list = []

    # for each single link id
    for single_id in df_link_dict:
        # define a flag to indicate whether an intersection exists
        isIntersection = False

        direction_list = df_link_dict[single_id]["Name"]
        direction_name_list = []

        # for link table, direction info store in column 3 to 10
        for direction_id in range(3, 11):
            direction_name = direction_list.get(utdf_link_column_names.get(direction_id), "")
            if direction_name not in direction_name_list and direction_name != '' and direction_name != '\n':
                direction_name_list.append(direction_name)
        if len(direction_name_list) > 1:
            isIntersection = True

        # generate intersection name if it's an intersection
        intersection_name = ""
        if isIntersection:
            intersection_name = ' & '.join(direction_name_list)
            # get "INTID"
            intersection_id = direction_list[utdf_link_column_names.get(2)]

            link_list.append([intersection_name, city_name, intersection_id, "", sequenced_intersection_id])
            sequenced_intersection_id += 1

    intersection_column_name = ["intersection_name", "city_name",
                                "synchro_INTID", "file_name", "intersection_id"]
    df_utdf_intersection = pd.DataFrame(link_list, columns=intersection_column_name)
    df_utdf_intersection["intersection_name"] = df_utdf_intersection["intersection_name"].map(lambda x: x.replace("\n", ""))

    return df_utdf_intersection


# combine phase and timeplans data into one dataframe (intersection id based)
def spanning_phase_timeplans_data(utdf_dict_data: dict, isSimpleCol: bool = True) -> pd.DataFrame:

    df_phase = utdf_dict_data.get("Phases")
    df_timeplans = utdf_dict_data.get("Timeplans")

    # get unique intersection id
    intersection_id = df_phase["INTID"].unique().tolist()

    final_spanned_list = []

    for INTID in intersection_id:

        # get utdf_phase dataframe by id
        df_phase_single_id = df_phase[df_phase["INTID"] == INTID].reset_index(drop=True)
        df_timeplans_single_id = df_timeplans[df_timeplans["INTID"] == INTID].reset_index(drop=True)

        df_phase_single_id_dict = df_phase_single_id.to_dict("list")
        df_timeplans_single_id_dict = df_timeplans_single_id.to_dict("list")

        # df phase new column name and data
        df_phase_col_name_new = []
        df_phase_single_id_new = []

        # df timeplans new column name and data
        df_timeplans_col_name_new = []
        df_timeplans_single_id_new = []

        # span df_phase_single_id_dict
        for col_name in df_phase_single_id_dict:
            if col_name not in ["RECORDNAME", "INTID"]:
                df_phase_col_name_new += [f"{row_name}_{col_name}" for row_name in df_phase_single_id_dict.get("RECORDNAME")]
                df_phase_single_id_new += df_phase_single_id_dict.get(col_name)

        # span df_timeplans_single_id_dict
        df_timeplans_col_name_new = df_timeplans_single_id_dict.get("RECORDNAME")
        df_timeplans_single_id_new = df_timeplans_single_id_dict.get("DATA")

        # generate final column name and data for single intersection id
        single_id_col_name_final = ["INTID"] + df_phase_col_name_new + df_timeplans_col_name_new
        single_id_data_final = [INTID] + df_phase_single_id_new + df_timeplans_single_id_new

        final_spanned_list.append(pd.DataFrame([single_id_data_final], columns=single_id_col_name_final))

        # if isSimpleCol is True, then only keep the columns with Start_D and End_D if _D exists
        # if _D not exists in column name, then keep all columns
        if isSimpleCol:
            simple_col_list = []
            for i in single_id_col_name_final:
                if "_D" in i:
                    if ("Start" in i and "Local" not in i) or "End" in i:
                        simple_col_list.append(i)
                else:
                    simple_col_list.append(i)
            final_spanned_list = [df[simple_col_list] for df in final_spanned_list]
            return pd.concat(final_spanned_list, axis=0, ignore_index=True)
    return pd.concat(final_spanned_list, axis=0, ignore_index=True)


if __name__ == '__main__':

    path_project_folder = Path(__file__).parents[2].absolute()
    path_utdf = path2linux(os.path.join(path_project_folder, "datasets/data_ASU_network_2/UTDF.csv"))

    # city_name = utdf_setting.get("city_name")

    utdf_dict_data = read_UTDF_file(path_utdf)

    df_phase = utdf_dict_data.get("Phases")
    df_timeplans = utdf_dict_data.get("Timeplans")

    intersection_id = df_phase["INTID"].unique().tolist()

    df_phase_spanned = spanning_phase_timeplans_data(utdf_dict_data)














# -*- coding:utf-8 -*-
##############################################################
# Created Date: Friday, January 27th 2023
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################

import os
import pickle

import pandas as pd

# For deployment
from utdf2gmns.func_lib.geocoding_intersection import \
    generate_coordinates_from_intersection
from utdf2gmns.func_lib.match_node_intersection_movement_utdf import (
    match_intersection_node, match_movement_and_intersection_node,
    match_movement_utdf_lane, match_movement_utdf_phase_timeplans)
from utdf2gmns.func_lib.read_utdf import (generate_intersection_data_from_utdf,
                                          read_UTDF_file)
from utdf2gmns.utils_lib.package_settings import (required_files,
                                                  required_files_sub,
                                                  utdf_city_name)
from utdf2gmns.utils_lib.utility_lib import (check_required_files_exist,
                                             func_running_time,
                                             get_filenames_from_folder_by_type,
                                             path2linux, validate_filename)

pd.options.mode.chained_assignment = None  # default='warn'


@func_running_time
def generate_utdf_dataframes(utdf_filename: str, city_name: str) -> dict:
    # read single UTDF file and produce data conversion and store data into a dictionary

    # read UTDF file and create dataframes of utdf_geo and utdf_lane
    utdf_dict_data = read_UTDF_file(utdf_filename)
    df_utdf_intersection = generate_intersection_data_from_utdf(utdf_dict_data, city_name)
    utdf_dict_data["utdf_intersection"] = df_utdf_intersection

    # geocoding utdf_intersection
    # df_utdf_geo = generate_coordinates_from_intersection(df_utdf_intersection)

    # store utdf_geo and utdf_lane into utdf_dict_data
    # utdf_dict_data["utdf_geo"] = df_utdf_geo

    # store object into pickle file
    # with open(os.path.join(os.getcwd(), "utdf2gmns.pickle"), 'wb') as f:
    #     pickle.dump(utdf_dict_data, f, pickle.HIGHEST_PROTOCOL)

    return utdf_dict_data


@func_running_time
def generate_movement_utdf(input_dir: str,
                           city_name: str = "",
                           output_dir: str = "",
                           isSave2csv: bool = True) -> list:

    # check if required files exist in the input directory
    files_from_directory = get_filenames_from_folder_by_type(input_dir, file_type="csv")

    # if not required, raise an exception
    isRequired = check_required_files_exist(required_files, files_from_directory)
    isRequired_sub = check_required_files_exist(required_files_sub, files_from_directory)

    # required files are not found, raise an exception
    if not isRequired:
        raise Exception(f"Required files {required_files} are not found!")

    # read UTDF file and create dataframes of utdf_intersection and utdf_lane
    path_utdf = path2linux(os.path.join(input_dir, "UTDF.csv"))

    if not city_name:
        print(f"City name is not provided, use default city name {utdf_city_name}")
        print("If you use a different city name, please provide it as a parameter")
        city_name = utdf_city_name

    utdf_dict_data = generate_utdf_dataframes(path_utdf, city_name)

    # geocoding the utdf_intersection and store it into utdf_dict_data
    # If not automatically geocode, save the utdf_geo.csv file in the input directory
    # And user manually add coord_x and coord_y to in utdf_geo.csv file
    # And re-run the function to generate the final movement_utdf.csv file

    # check utdf_geo.csv file existence
    if os.path.exists(path2linux(os.path.join(input_dir, "utdf_geo.csv"))):
        # read utdf_geo.csv file from the input directory
        utdf_dict_data["utdf_geo"] = pd.read_csv(path2linux(
            os.path.join(input_dir, "utdf_geo.csv")))

        # check if user manually added coord_x and coord_y to in utdf_geo.csv file
        if not {"coord_x", "coord_y"}.issubset(set(utdf_dict_data.get("utdf_geo").columns)):
            raise Exception(
                "coord_x or coord_y not found in the utdf_geo.csv file!, please add coord_x and coord_y manually \
                 and re-run the code afterwards."
            )

    # check utdf_geo in utdf_dict_data or not
    # if not generate utdf_geo automatically
    if "utdf_geo" not in utdf_dict_data.keys():
        try:
            # geocoding utdf_intersection automatically
            utdf_dict_data["utdf_geo"] = generate_coordinates_from_intersection(
                utdf_dict_data.get("utdf_intersection"))
        except Exception as e:
            # #  Save the utdf_geo.csv file in the input directory
            utdf_dict_data.get("utdf_intersection").to_csv(
                path2linux(os.path.join(input_dir, "utdf_geo.csv")), index=False)

            raise Exception(
                "We can not geocoding intersections automatically, \
                 We save utdf_geo.csv file in your input dir,   \
                please manually add coord_x and coord_y to the utdf_geo.csv file \
                in your input directory and re-run the code afterwards."
            ) from e

    # required_sub files are not found, will return utdf_intersection and utdf_lane
    if not isRequired_sub:
        print("Because node.csv and movement.csv are not found, \
            the function will return data from utdf in a dictionary, \
            keys are: Lanes, Nodes, Networks, Timeplans, Links and utdf_geo.\n")

        # store object into pickle file
        with open(path2linux(os.path.join(input_dir, "utdf2gmns.pickle")), 'wb') as f:
            pickle.dump(utdf_dict_data, f, pickle.HIGHEST_PROTOCOL)
        return utdf_dict_data

    # get the path of each file,
    # since the input directory and files are checked, no need to validate the filename
    path_node = path2linux(os.path.join(input_dir, "node.csv"))
    path_movement = path2linux(os.path.join(input_dir, "movement.csv"))

    # read node and movement files
    df_node = pd.read_csv(path_node)
    df_movement = pd.read_csv(path_movement)

    # match utdf_intersection_geo with node
    print("Performing data matching between df_utdf_intersection_geo and node...")
    df_intersection_node = match_intersection_node(utdf_dict_data.get("utdf_geo"),
                                                   df_node)

    # match movement with intersection_node
    print("Performing data matching between movement and intersection_node...")
    df_movement_intersection = match_movement_and_intersection_node(df_movement,
                                                                    df_intersection_node)

    # match movement with utdf_lane
    print("Performing data matching between movement_geo and utdf_lane...")
    df_movement_utdf_lane = match_movement_utdf_lane(
        df_movement_intersection, utdf_dict_data)

    # match movement with utdf_phase_timeplans
    print("Performing data matching between movement_utdf_lane and utdf_phase_timeplans...")
    df_movement_utdf_phase = match_movement_utdf_phase_timeplans(
        df_movement_utdf_lane, utdf_dict_data)

    # store utdf_intersection_geo and movement_utdf to utdf_dict_data
    utdf_dict_data["movement_utdf_phase"] = df_movement_utdf_phase
    utdf_dict_data["utdf_geo_GMNS_node"] = df_intersection_node

    # save the output file, the default isSave2csv is True
    # if not specified, output path is input directory,
    # output file name = movement_utdf.csv
    if isSave2csv:
        if not output_dir:
            output_dir = input_dir
        output_file_name = validate_filename(os.path.join(output_dir, "movement_utdf.csv"))
        df_movement_utdf_phase.to_csv(output_file_name, index=False)

        output_file_name = validate_filename(os.path.join(output_dir, "utdf_intersection.csv"))
        df_intersection_node.to_csv(output_file_name, index=False)

        # with open(path2linux(os.path.join(output_dir, "utdf2gmns.pickle")), 'wb') as f:
        #     pickle.dump(utdf_dict_data, f, pickle.HIGHEST_PROTOCOL)

    return [df_movement_utdf_phase, utdf_dict_data]


if __name__ == '__main__':

    city_name = " Bullhead City, AZ"

    # NOTE : the following code is for generating movement_utdf.csv file
    input_dir = r"C:\Users\roche\Anaconda_workspace\001_Github\utdf2gmns\datasets\data_bullhead_seg4_1"
    df_movement_utdf_phase, utdf_dict_data = generate_movement_utdf(input_dir, city_name, isSave2csv=False)
    # df_movement_utdf_phase.to_csv(path2linux(os.path.join(input_dir, "movement_utdf.csv")), index=False)

    # # NOTE : the following code is for testing purpose only: read utdf.csv file and generate utdf_dict_data
    # path_utdf = r"C:\Users\roche\Anaconda_workspace\001_Github\utdf2gmns\datasets\data_ASU_network\UTDF.csv"
    # utdf_dict_data = generate_utdf_dataframes(path_utdf, city_name)



# -*- coding:utf-8 -*-
##############################################################
# Created Date: Tuesday, January 31st 2023
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################

from .__utdf2gmns import (generate_utdf_dataframes,
                          generate_movement_utdf,
                          generate_coordinates_from_intersection,
                          match_intersection_node,
                          match_movement_and_intersection_node,
                          match_movement_utdf_lane,
                          match_movement_utdf_phase_timeplans
                          )

from .utils_lib import package_settings

__all__ = ['generate_utdf_dataframes',
           'generate_movement_utdf',
           'generate_coordinates_from_intersection',
           'match_intersection_node',
           'match_movement_and_intersection_node',
           'match_movement_utdf_lane',
           'match_movement_utdf_phase_timeplans',
           'package_settings'
           ]

print("UTDF2GMNS Version: ", "0.2.1")


# -*- coding:utf-8 -*-
##############################################################
# Created Date: Wednesday, November 16th 2022
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################

from math import sin, cos, sqrt, atan2, radians
import os
import datetime
from pathlib import Path
from typing import Union  # Python version <= 3.9


# A decorator to measure the time of a function
def func_running_time(func):
    def inner(*args, **kwargs):
        print(f'INFO Begin to run function: {func.__name__} …')
        time_start = datetime.datetime.now()
        res = func(*args, **kwargs)
        time_diff = datetime.datetime.now() - time_start
        print(
            f'INFO Finished running function: {func.__name__}, total: {time_diff.seconds}s')
        print()
        return res
    return inner


# convert OS path to standard linux path
def path2linux(path: Union[str, Path]) -> str:
    """Convert a path to a linux path, linux path can run in windows, linux and mac"""
    try:
        return path.replace("\\", "/")
    except Exception:
        return str(path).replace("\\", "/")


def get_filenames_from_folder_by_type(dir_name: str, file_type: str = "txt", isTraverseSubdirectory: bool = False) -> list:
    """Get all files in the folder with the specified file type

    Args:
        dir_name (str)                         : the folder path
        file_type (str, optional)              : the exact file type to specify, if file_type is "*" or "all", return all files in the folder. Defaults to "txt".
        isTraverseSubdirectory (bool, optional): get files inside the subfolder or not, if True, will traverse all subfolders. Defaults to False.

    Returns:
        list: a list of file paths

    Examples:
        # get all files in the folder without traversing subfolder
        >>> from pyhelpers.dirs import get_filenames_from_folder_by_type
        >>> get_filenames_from_folder_by_type("C:/Users/user/Desktop", "txt")
        ['C:/Users/user/Desktop/test.txt']

        # get all files in the folder with traversing subfolder
        >>> from pyhelpers.dirs import get_filenames_from_folder_by_type
        >>> get_filenames_from_folder_by_type("C:/Users/user/Desktop", "txt", isTraverseSubdirectory=True)
        ['C:/Users/user/Desktop/test.txt', 'C:/Users/user/Desktop/sub_folder/test2.txt']
    """

    if isTraverseSubdirectory:
        files_list = []
        for root, dirs, files in os.walk(dir_name):
            files_list.extend([os.path.join(root, file) for file in files])
        if file_type in {"*", "all"}:
            return [path2linux(file) for file in files_list]
        return [path2linux(file) for file in files_list if file.split(".")[-1] == file_type]
    print("input dir:", dir_name, "input file type", file_type)
    # files in the first layer of the folder
    if file_type in {"*", "all"}:
        return [path2linux(os.path.join(dir_name, file)) for file in os.listdir(dir_name)]
    return [path2linux(os.path.join(dir_name, file)) for file in os.listdir(dir_name) if file.split(".")[-1] == file_type]


def check_required_files_exist(required_files: list, dir_files: list) -> bool:
    # format the required file name to standard linux path
    required_files = [path2linux(os.path.abspath(filename)) for filename in required_files]

    required_files_short = [filename.split("/")[-1] for filename in required_files]
    dir_files_short = [filename.split("/")[-1] for filename in dir_files]

    # mask have the same length as required_files
    mask = [file in dir_files_short for file in required_files_short]
    if all(mask):
        return True

    print(f"Error: Required files are not satisfied, \
          missing files are: {[required_files_short[i] for i in range(len(required_files_short)) if not mask[i]]}")

    return False


def validate_filename(path_filename: str, ) -> str:
    """if the file name exist in path,then create new file name with _1, _1_1, ..."""
    filename_abspath = path2linux(os.path.abspath(path_filename))

    file_suffix = filename_abspath.split(".")[-1]
    file_without_suffix = filename_abspath[:-len(file_suffix) - 1]

    if os.path.exists(filename_abspath):
        filename_update = f"{file_without_suffix}_1.{file_suffix}"
        return validate_filename(filename_update)
    return filename_abspath


def calculate_point2point_distance_in_km(point1: tuple, point2: tuple) -> float:
    """ point1 and point2: a tuple of (longitude, latitude) """

    # approximate radius of earth in km
    R = 6373.0

    lat1 = radians(point1[1])
    lon1 = radians(point1[0])
    lat2 = radians(point2[1])
    lon2 = radians(point2[0])

    lon_diff = lon2 - lon1
    lat_diff = lat2 - lat1

    a = sin(lat_diff / 2)**2 + cos(lat1) * cos(lat2) * sin(lon_diff / 2)**2
    c = 2 * atan2(sqrt(a), sqrt(1 - a))

    distance = R * c

    # return math.pow((point1[0] - point2[0]) ** 2 + (point1[1] - point2[1]) ** 2, 0.5)

    # the distance is in km
    return distance


if __name__ == "__main__":
    dir_name = r"C:\Users\roche\Anaconda_workspace\001_Github.com\GTFS2GMNS\test\GTFS"

    # files_from_folder_abspath = get_txt_files_from_folder(dir_name, isTraverseSubdirectory=False)

    # files_required = ["agency.txt", "routes.txt", "shapes.txt", "stops.txt", "trips.txt"]
    # required_files_abspath = [path2linux(os.path.join(dir_name, file)) for file in files_required]

    # isFilesExist = check_required_files_exist(required_files_abspath, files_from_folder_abspath)

    filename = "./node.csv"
    ss = validate_filename("./node.csv")



# -*- coding:utf-8 -*-
##############################################################
# Created Date: Monday, November 28th 2022
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################

# from geopy.geocoders import Nominatim
# import googlemaps
import contextlib
# import os
# import sys
from multiprocessing import Pool
# from pathlib import Path

import geocoder
import pandas as pd

from utdf2gmns.utils_lib.utility_lib import (
    calculate_point2point_distance_in_km, func_running_time)

# def googlemaps_geocoding_from_address(address, api_key) -> tuple:
#
#     # initialize googlemaps client
#     gmaps = googlemaps.Client(key=api_key)
#
#     # Geocoding an address
#     location_instance = gmaps.geocode(address)
#
#     # get the location
#     location_lng_lat = (location_instance[0]['geometry']['location']['lat'], location_instance[0]['geometry']['location']['lng'])
#
#     return location_lng_lat
#
#
# def geopy_geocoding_from_address(address: str) -> tuple:
#
#     # initialize geopy client
#     geo_locator = Nominatim(user_agent="myGeopyGeocoder")
#
#     # Geocoding an address
#     try:
#         location = geo_locator.geocode(address, timeout=10)
#         location_lng_lat = (location.longitude, location.latitude)
#     except Exception as e:
#
#         location_lng_lat = (0,0)
#         print(
#             f"Error: {address} is not able to geocode, for {e}, try to use (0 ,0) to as lng and lat \n")
#
#     return location_lng_lat


def geocoder_geocoding_from_address(address: str) -> tuple:

    # initialize geocoder arcgis client
    location_instance = geocoder.arcgis(address).geojson

    # get the location
    try:
        location_lng_lat = (
            location_instance["features"][0]["geometry"]["coordinates"])
    except Exception:
        location_lng_lat = [0, 0]

    return location_lng_lat


@func_running_time
def generate_coordinates_from_intersection(df_intersection: pd.DataFrame, distance_threshold=0.01) -> pd.DataFrame:
    # distance_threshold is the threshold to determine whether the intersection is able to geocode, using km as unit

    df = df_intersection.copy()

    # check required columns exist in the dataframe
    if not {"intersection_name", "city_name"}.issubset(set(df.columns)):
        raise Exception(
            "intersection_name and city_name are not in the dataframe, please check the input file.")

    # Create one column named "reversed_intersection_name"
    for i in range(len(df)):
        if "&" in df.loc[i, "intersection_name"]:
            df.loc[i, "reversed_intersection_name"] = df.loc[i, "intersection_name"].split(
                "&")[1] + " & " + df.loc[i, "intersection_name"].split(" & ")[0]
        else:
            df.loc[i, "reversed_intersection_name"] = df.loc[i, "intersection_name"]

    # create two columns named "full_name_intersection" and "full_name_intersection_reversed"
    df["full_name_intersection"] = df["intersection_name"] + ", " + df["city_name"]
    df["full_name_intersection_reversed"] = df["reversed_intersection_name"] + ", " + df["city_name"]

    # Step 4: geocoding
    print("   :geocoding intersections...")
    intersection_full_name_list = df["full_name_intersection"].tolist()
    intersection_full_name_reversed_list = df["full_name_intersection_reversed"].tolist()

    try:
        # use multiprocessing to speed up
        # p = Pool()
        # lnglat_values_full_name = p.map(geocoder_geocoding_from_address, intersection_full_name_list)

        # using loop to geocode
        lnglat_values_full_name = [geocoder_geocoding_from_address(address) for address in intersection_full_name_list]

    except Exception as e:
        raise Exception("   :geocoding intersections failed, try again...") from e

    try:
        # use multiprocessing to speed up
        # p1 = Pool()
        # lnglat_values_full_name_reversed = p1.map(geocoder_geocoding_from_address, intersection_full_name_reversed_list)

        # using loop to geocode
        lnglat_values_full_name_reversed = [geocoder_geocoding_from_address(address) for address in intersection_full_name_reversed_list]

    except Exception as exc:
        raise Exception("   :geocoding intersections failed, try again...") from exc

    # create new column named distance_from_full_name
    print("   :cross validating...")
    distance = [calculate_point2point_distance_in_km(
        lnglat_values_full_name[i], lnglat_values_full_name_reversed[i]) for i in range(len(lnglat_values_full_name))]

    for i in range(len(df)):
        df["distance_from_full_name"] = distance

        if distance[i] <= distance_threshold:
            df.loc[i, "coord_x"] = lnglat_values_full_name[i][0]
            df.loc[i, "coord_y"] = lnglat_values_full_name[i][1]

        else:
            # use None to indicate the intersection is not able to geocode
            df.loc[i, "coord_x"] = 0
            df.loc[i, "coord_y"] = 0

    created_column_names = ["reversed_intersection_name", "full_name_intersection",
                            "full_name_intersection_reversed", "distance_from_full_name"]
    df_final = df.loc[:, ~df.columns.isin(created_column_names)]

    # print summary information
    print(
        f" {len(df_final) - df_final['coord_x'].isna().sum()} / {len(df_final)} intersections geocoded.")
    print(
        f" {df_final['coord_x'].isna().sum()} / {len(df_final)} intersections are not able to geocode.")

    return df_final


if __name__ == "__main__":
    """
    We provided three geocoding methods: googlemaps, geopy and geocoder
    The example using geocoder and you don't need to install googlemaps and geopy as well(comment out the code if you want to use them)
    """

    # Step 1: input data path
    path_input = r"C:\Users\roche\Anaconda_workspace\001_Github\utdf2gmns\datasets\data_test_1\utdf_intersection.csv"

    # Step 2: read data
    df_intersection = pd.read_csv(path_input)

    # Step 2: generate coordinates
    df = generate_coordinates_from_intersection(df_intersection)

    # save to csv file
    # df.to_csv(f"{path_input.split('.')[0]}_with_coordinates.csv", index=False)
